#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2024 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of Pandora plugin LibSGM
#
#     https://github.com/CNES/Pandora_plugin_libsgm
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module provides class and functions to compute penalties used to optimize the cost volume using the LibSGM library
"""

from typing import Dict, Union, Tuple, List

import numpy as np
import xarray as xr
from json_checker import Checker, And, Or

from pandora.common import is_method
from pandora_plugin_libsgm.penalty import penalty


@penalty.AbstractPenalty.register_subclass("mc_cnn_fast_penalty")
class MccnnPenalty(penalty.AbstractPenalty):
    """

    MC-CNN Penalty

    """

    # Default mc-cnn fast penalty
    _P1 = 2.3
    _P2 = 42.3
    _Q1 = 3
    _Q2 = 6
    _D = 0.08
    _V = 1.5

    def __init__(self, directions: List[List[int]], **cfg: Union[str, int, float, bool]):
        """
        :param directions: directions to
        :type directions: list of [x offset, y offset]
        :param cfg: optional configuration, {'P1': value, 'P2': value, 'Q1': value, 'Q2': value, 'D": value,
                                            'V': value}
        :type cfg: dict
        """
        self.cfg = self.check_conf(**cfg)
        self._p1 = self.cfg["P1"]
        self._p2 = self.cfg["P2"]
        self._q1 = self.cfg["Q1"]
        self._q2 = self.cfg["Q2"]
        self._d = self.cfg["D"]
        self._v = self.cfg["V"]
        self._directions = directions

    def check_conf(self, **cfg: Union[str, int, float, bool]) -> Dict[str, Union[str, int, float, bool]]:
        """
        Add default values to the dictionary if there are missing elements and check if the dictionary is correct

        :param cfg: optimization configuration
        :type cfg: dict
        :return cfg: optimization configuration updated
        :rtype cfg: dict
        """

        # Give the default value if the required element is not in the configuration
        if "P1" not in cfg:
            cfg["P1"] = self._P1
        if "P2" not in cfg:
            cfg["P2"] = self._P2
        if "Q1" not in cfg:
            cfg["Q1"] = self._Q1
        if "Q2" not in cfg:
            cfg["Q2"] = self._Q2
        if "D" not in cfg:
            cfg["D"] = self._D
        if "V" not in cfg:
            cfg["V"] = self._V

        p1_value = cfg["P1"]
        schema = {
            "penalty_method": And(
                str,
                lambda x: is_method(x, ["mc_cnn_fast_penalty"]),
            ),
            "P1": And(Or(int, float), lambda x: x > 0),
            "P2": And(Or(int, float), lambda x: x > p1_value),
            "Q1": And(Or(int, float), lambda x: x > 0),
            "Q2": And(Or(int, float), lambda x: x > 0),
            "D": And(Or(int, float), lambda x: x >= 0),
            "V": And(Or(int, float), lambda x: x > 0),
        }

        checker = Checker(schema)
        checker.validate(cfg)
        return cfg

    def desc(self):
        """
        Describes the penalty method

        """
        print("MC-CNN penalty method description")

    def compute_penalty(
        self, cv: xr.Dataset, img_left_array: np.ndarray, img_right_array: np.ndarray
    ) -> Tuple[float, np.ndarray, np.ndarray]:
        """
        Compute penalty

        :param cv: the cost volume, with the data variables:

            - cost_volume 3D xarray.DataArray (row, col, disp)
            - confidence_measure 3D xarray.DataArray (row, col, indicator)
        :type cv: xarray.Dataset
        :param img_left_array: left  image
        :type img_left_array: np.ndarray
        :param img_right_array: right  image
        :type img_right_array: np.ndarray
        :return: invalid disp, P1 and P2 penalties
        :rtype: tuple(float, numpy array, numpy array)
        """

        # Calculation of the invalid value
        p2_max = max(self._p2, self._p2 / self._q2, self._p2 / self._p1)  # type: ignore
        invalid_value = float(cv.attrs["cmax"] + p2_max + 1)

        # Compute penalties

        p1_mask, p2_mask = self.mc_cnn_penalty_function(
            img_left_array,
            img_right_array,
            self._p1,  # type: ignore
            self._p2,  # type: ignore
            self._q1,  # type: ignore
            self._q2,  # type: ignore
            self._d,  # type: ignore
            self._v,  # type: ignore
            self._directions,
        )

        return invalid_value, p1_mask, p2_mask

    def mc_cnn_penalty_function(  # pylint: disable=too-many-arguments
        self,
        img_left: np.ndarray,
        img_right: np.ndarray,
        p1_mccnn: Union[int, float],
        p2_mccnn: Union[int, float],
        q1_mccnn: Union[int, float],
        q2_mccnn: Union[int, float],
        d_mccnn: Union[int, float],
        v_mccnn: Union[int, float],
        directions: List[List[int]],
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Compute mc_cnn penalty

        :param img_left: left  image
        :type img_left: numpy array
        :param img_right: right  image
        :type img_right: numpy array
        :param p1_mccnn:  P1 penalty
        :type p1_mccnn: int or float
        :param p2_mccnn: default P2 penalty
        :type p2_mccnn: int or float
        :param q1_mccnn: hyper parameter
        :type q1_mccnn: int or float
        :param q2_mccnn: hyper parameter
        :type q2_mccnn: int or float
        :param d_mccnn: hyper parameter
        :type d_mccnn: int or float
        :param v_mccnn: hyper parameter
        :type v_mccnn: int or float
        :param directions: directions to
        :type directions: list of [x offset, y offset]
        :return: P1 and P2 penalties
        :rtype: tuple(numpy array, numpy array)
        """
        p1_mask = p1_mccnn * np.ones([img_left.shape[0], img_left.shape[1], len(directions)], dtype=np.float32)
        p2_mask = p2_mccnn * np.ones([img_left.shape[0], img_left.shape[1], len(directions)], dtype=np.float32)

        nb_directions = len(directions)
        for i in range(nb_directions):
            direction = directions[i]
            abs_gradient_left = self.compute_gradient(img_left[:, :], direction)
            abs_gradient_right = self.compute_gradient(img_right[:, :], direction)
            # if(D1<sgm_D && D2<sgm_D)
            msk1 = (abs_gradient_left < d_mccnn) * (abs_gradient_right < d_mccnn)
            final_p1 = msk1 * p1_mccnn * np.ones(abs_gradient_left.shape)
            final_p2 = msk1 * p2_mccnn * np.ones(abs_gradient_left.shape)
            # if(D1 > sgm_D && D2 > sgm_D)
            msk2 = (abs_gradient_left > d_mccnn) * (abs_gradient_right > d_mccnn)
            final_p1 = final_p1 + msk2 * (p1_mccnn / (q1_mccnn * q2_mccnn)) * np.ones(abs_gradient_left.shape)
            final_p2 = final_p2 + msk2 * (p2_mccnn / (q1_mccnn * q2_mccnn)) * np.ones(abs_gradient_left.shape)
            # else
            msk3 = (1 - msk1) * (1 - msk2)
            final_p1 = final_p1 + msk3 * (p1_mccnn / q1_mccnn) * np.ones(abs_gradient_left.shape)
            final_p2 = final_p2 + msk3 * (p2_mccnn / q1_mccnn) * np.ones(abs_gradient_left.shape)

            p1_mask[
                max(0, direction[0]) : min(img_left.shape[0] + direction[0], img_left.shape[0]),
                max(0, direction[1]) : min(img_left.shape[1] + direction[1], img_left.shape[1]),
                i,
            ] = final_p1
            p2_mask[
                max(0, direction[0]) : min(img_left.shape[0] + direction[0], img_left.shape[0]),
                max(0, direction[1]) : min(img_left.shape[1] + direction[1], img_left.shape[1]),
                i,
            ] = final_p2

            if i in [1, 5]:
                p1_mask[:, :, i] = p1_mask[:, :, i] / v_mccnn

        return p1_mask, p2_mask



#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2024 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of Pandora plugin LibSGM
#
#     https://github.com/CNES/Pandora_plugin_libsgm
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module provides functions to test sgm penalties
"""

# pylint: disable=redefined-outer-name

import numpy as np
import pytest

from pandora_plugin_libsgm.penalty import penalty_sgm


@pytest.fixture()
def segment_penalty():
    """Segment penalty fixture."""
    cfg = {
        "P1": 8,
        "P2": 10,
        "alpha": 1.0,
        "gamma": 1,
        "beta": 1,
        "p2_method": "constant",
        "penalty_method": "sgm_penalty",
    }

    directions = [
        [0, 1],
        [1, 0],
        [1, 1],
        [1, -1],
        [0, -1],
        [-1, 0],
        [-1, -1],
        [-1, 1],
    ]

    return penalty_sgm.SgmPenalty(directions, **cfg)


class TestPenalitySGM:
    """
    TestPenalitySGM class allows to test penality_sgm
    """

    def test_gradient(self, segment_penalty):
        """
        Test Computation of gradient

        """

        img_left = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        # TEST 1
        img_wanted = np.array([[3, 3, 3], [3, 3, 3]])

        direction = [1, 0]
        computed_gradient = segment_penalty.compute_gradient(img_left, direction)
        # Check if the calculated gradient is equal to the ground truth (same shape and all elements equals)
        np.testing.assert_array_equal(computed_gradient, img_wanted)

        # TEST 2
        img_wanted = np.array([[1, 1], [1, 1], [1, 1]])

        direction = [0, 1]
        computed_gradient = segment_penalty.compute_gradient(img_left, direction)
        # Check if the calculated gradient is equal to the ground truth (same shape and all elements equals)
        np.testing.assert_array_equal(computed_gradient, img_wanted)

        # TEST 3
        img_wanted = np.array([[4, 4], [4, 4]])

        direction = [1, 1]
        computed_gradient = segment_penalty.compute_gradient(img_left, direction)
        # Check if the calculated gradient is equal to the ground truth (same shape and all elements equals)
        np.testing.assert_array_equal(computed_gradient, img_wanted)

        # TEST 4
        img_wanted = np.array([[4, 4], [4, 4]])

        direction = [-1, -1]
        computed_gradient = segment_penalty.compute_gradient(img_left, direction)
        # Check if the calculated gradient is equal to the ground truth (same shape and all elements equals)
        np.testing.assert_array_equal(computed_gradient, img_wanted)

        # TEST 5
        img_wanted = np.array([[2, 2], [2, 2]])

        direction = [-1, 1]
        computed_gradient = segment_penalty.compute_gradient(img_left, direction)
        # Check if the calculated gradient is equal to the ground truth (same shape and all elements equals)
        np.testing.assert_array_equal(computed_gradient, img_wanted)

    def test_constant_penalty_function(self, segment_penalty):
        """
        Test Computation of gradient

        """
        default_p1 = 1
        default_p2 = 2

        img_left = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        # TEST 1
        p1_wanted = np.array(
            [
                [[1, 1, 1], [1, 1, 1], [1, 1, 1]],
                [[1, 1, 1], [1, 1, 1], [1, 1, 1]],
                [[1, 1, 1], [1, 1, 1], [1, 1, 1]],
            ]
        )

        p2_wanted = np.array(
            [
                [[2, 2, 2], [2, 2, 2], [2, 2, 2]],
                [[2, 2, 2], [2, 2, 2], [2, 2, 2]],
                [[2, 2, 2], [2, 2, 2], [2, 2, 2]],
            ]
        )

        directions = [[1, 0], [-1, 1], [1, 1]]
        computed_p1, computed_p2 = segment_penalty.constant_penalty_function(
            img_left, default_p1, default_p2, directions
        )
        # Check if the calculated gradient is equal to the ground truth (same shape and all elements equals)
        np.testing.assert_array_equal(computed_p1, p1_wanted)
        np.testing.assert_array_equal(computed_p2, p2_wanted)

    def test_inverse_penalty_function(self, segment_penalty):
        """
        Test Computation of gradient

        """
        default_p1 = 1
        default_p2 = 2

        img_left = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        # TEST 1
        p1_wanted = np.array(
            [
                [[1, 1, 1], [1, 1, 1], [1, 1, 1]],
                [[1, 1, 1], [1, 1, 1], [1, 1, 1]],
                [[1, 1, 1], [1, 1, 1], [1, 1, 1]],
            ]
        )

        p2_wanted = np.array(
            [
                [[2, 2, 2], [2, 2, 2], [2, 2, 2]],
                [[2, 2, 2], [2, 2, 2], [2, 2, 2]],
                [[2, 2, 2], [2, 2, 2], [2, 2, 2]],
            ]
        )

        directions = [[1, 0], [-1, 1], [1, 1]]
        alpha = 1
        beta = 1
        gamma = 1
        computed_p1, computed_p2 = segment_penalty.inverse_penalty_function(
            img_left, default_p1, default_p2, directions, alpha, beta, gamma
        )
        # Check if the calculated gradient is equal to the ground truth (same shape and all elements equals)
        np.testing.assert_array_equal(computed_p1, p1_wanted)
        np.testing.assert_array_equal(computed_p2, p2_wanted)

        # TEST 2

        default_p1 = 1
        default_p2 = 0

        img_left = np.array(
            [
                [1, 2, 3, 4, 5],
                [6, 7, 8, 9, 10],
                [11, 12, 13, 14, 15],
                [16, 17, 18, 19, 20],
                [21, 22, 23, 24, 25],
            ]
        )
        p1_wanted_0 = np.array(
            [
                [1, 1, 1, 1, 1],
                [1, 1, 1, 1, 1],
                [1, 1, 1, 1, 1],
                [1, 1, 1, 1, 1],
                [1, 1, 1, 1, 1],
            ]
        )

        p2_wanted_0 = np.array(
            [
                [0, 0, 0, 0, 0],
                [7 / 6, 7 / 6, 7 / 6, 7 / 6, 7 / 6],
                [7 / 6, 7 / 6, 7 / 6, 7 / 6, 7 / 6],
                [7 / 6, 7 / 6, 7 / 6, 7 / 6, 7 / 6],
                [7 / 6, 7 / 6, 7 / 6, 7 / 6, 7 / 6],
            ],
            dtype=np.float32,
        )

        p2_wanted_1 = np.array(
            [
                [0, 6 / 5, 6 / 5, 6 / 5, 6 / 5],
                [0, 6 / 5, 6 / 5, 6 / 5, 6 / 5],
                [0, 6 / 5, 6 / 5, 6 / 5, 6 / 5],
                [0, 6 / 5, 6 / 5, 6 / 5, 6 / 5],
                [0, 0, 0, 0, 0],
            ],
            dtype=np.float32,
        )

        p2_wanted_2 = np.array(
            [
                [0, 0, 0, 0, 0],
                [0, 8 / 7, 8 / 7, 8 / 7, 8 / 7],
                [0, 8 / 7, 8 / 7, 8 / 7, 8 / 7],
                [0, 8 / 7, 8 / 7, 8 / 7, 8 / 7],
                [0, 8 / 7, 8 / 7, 8 / 7, 8 / 7],
            ],
            dtype=np.float32,
        )

        directions = [[1, 0], [-1, 1], [1, 1]]
        alpha = 1
        beta = 1
        gamma = 1
        computed_p1, computed_p2 = segment_penalty.inverse_penalty_function(
            img_left, default_p1, default_p2, directions, alpha, beta, gamma
        )
        # Check if the calculated gradient is equal to the ground truth (same shape and all elements equals)
        np.testing.assert_array_equal(computed_p1[:, :, 0], p1_wanted_0)
        np.testing.assert_array_equal(computed_p2[:, :, 0], p2_wanted_0)
        np.testing.assert_array_equal(computed_p2[:, :, 1], p2_wanted_1)
        np.testing.assert_array_equal(computed_p2[:, :, 2], p2_wanted_2)

    def test_negative_penalty_function(self, segment_penalty):
        """
        Test Computation of gradient

        """
        default_p1 = 1
        default_p2 = 2

        img_left = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        # TEST 1
        p1_wanted = np.array(
            [
                [[1, 1, 1], [1, 1, 1], [1, 1, 1]],
                [[1, 1, 1], [1, 1, 1], [1, 1, 1]],
                [[1, 1, 1], [1, 1, 1], [1, 1, 1]],
            ]
        )

        p2_wanted = np.array(
            [
                [[2, 2, 2], [2, 2, 2], [2, 2, 2]],
                [[2, 2, 2], [2, 2, 2], [2, 2, 2]],
                [[2, 2, 2], [2, 2, 2], [2, 2, 2]],
            ]
        )

        directions = [[1, 0], [-1, 1], [1, 1]]
        alpha = 1
        gamma = 1
        computed_p1, computed_p2 = segment_penalty.negative_penalty_function(
            img_left, default_p1, default_p2, directions, alpha, gamma
        )
        # Check if the calculated gradient is equal to the ground truth (same shape and all elements equals)
        np.testing.assert_array_equal(computed_p1, p1_wanted)
        np.testing.assert_array_equal(computed_p2, p2_wanted)

        # TEST 2

        default_p1 = 1
        default_p2 = 0

        img_left = np.array(
            [
                [1, 2, 3, 4, 5],
                [6, 7, 8, 9, 10],
                [11, 12, 13, 14, 15],
                [16, 17, 18, 19, 20],
                [21, 22, 23, 24, 25],
            ]
        )
        p1_wanted_0 = np.array(
            [
                [1, 1, 1, 1, 1],
                [1, 1, 1, 1, 1],
                [1, 1, 1, 1, 1],
                [1, 1, 1, 1, 1],
                [1, 1, 1, 1, 1],
            ]
        )

        p2_wanted_0 = np.array(
            [
                [0, 0, 0, 0, 0],
                [5, 5, 5, 5, 5],
                [5, 5, 5, 5, 5],
                [5, 5, 5, 5, 5],
                [5, 5, 5, 5, 5],
            ]
        )

        p2_wanted_1 = np.array(
            [
                [0, 6, 6, 6, 6],
                [0, 6, 6, 6, 6],
                [0, 6, 6, 6, 6],
                [0, 6, 6, 6, 6],
                [0, 0, 0, 0, 0],
            ]
        )

        p2_wanted_2 = np.array(
            [
                [0, 0, 0, 0, 0],
                [0, 4, 4, 4, 4],
                [0, 4, 4, 4, 4],
                [0, 4, 4, 4, 4],
                [0, 4, 4, 4, 4],
            ]
        )

        directions = [[1, 0], [-1, 1], [1, 1]]
        alpha = 1
        gamma = 10
        computed_p1, computed_p2 = segment_penalty.negative_penalty_function(
            img_left, default_p1, default_p2, directions, alpha, gamma
        )
        # Check if the calculated gradient is equal to the ground truth (same shape and all elements equals)
        np.testing.assert_array_equal(computed_p1[:, :, 0], p1_wanted_0)
        np.testing.assert_array_equal(computed_p2[:, :, 0], p2_wanted_0)
        np.testing.assert_array_equal(computed_p2[:, :, 1], p2_wanted_1)
        np.testing.assert_array_equal(computed_p2[:, :, 2], p2_wanted_2)



# Copyright (c) 2024 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of Pandora plugin LibSGM
#
#     https://github.com/CNES/Pandora_plugin_libsgm
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains functions to test all plugin configurations.
"""

import pytest
import json_checker

from pandora import optimization

pytestmark = pytest.mark.usefixtures("import_plugin")

class TestCheckConf():

    @pytest.fixture()
    def optimization_cfg(self, use_confidence_value):
        return {"optimization_method": "sgm", "use_confidence": use_confidence_value}
    
    @pytest.mark.parametrize(
        "use_confidence_value",
        [
            pytest.param(1, id="Integer"),
            pytest.param([1, 1], id="Integer list"),
            pytest.param((1, 1), id="Integer Tuple"),
            pytest.param(True, id="Boolean"),
            pytest.param({"test": 1}, id="Dict"),
        ],
    )
    def test_with_wrong_use_confidence_parameter(self, left_crafted, optimization_cfg):
        with pytest.raises(json_checker.core.exceptions.DictCheckerError):
            optimization.AbstractOptimization(left_crafted, **optimization_cfg)

    @pytest.mark.parametrize("use_confidence_value", ["cost_volume_confidence"])
    def test_with_nominal_use_confidence_parameter(self, left_crafted, optimization_cfg):
        optimization.AbstractOptimization(left_crafted, **optimization_cfg)

    @pytest.mark.parametrize("use_confidence_value", ["cost_volume_confidence"])
    def test_without_use_confidence_parameter(self, left_crafted, optimization_cfg):
        del optimization_cfg["use_confidence"]
        optimization.AbstractOptimization(left_crafted, **optimization_cfg)



#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2024 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of Pandora plugin LibSGM
#
#     https://github.com/CNES/Pandora_plugin_libsgm
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module provides functions to optimize the cost volume using the 3SGM method

"""

import logging
import sys
from typing import Union, Tuple

import numpy as np
import xarray as xr
from pandora.img_tools import fuse_classification_bands
from . import abstract_sgm


@abstract_sgm.AbstractSGM.register_subclass("3sgm")
class SEGSEMSGM(abstract_sgm.AbstractSGM):
    """

    SEGSEMSGM class is a plugin that allow to perform the optimization step
    by calling the LibSGM library with an optimization layer.

    """

    # Default configuration, do not change these values
    _GEOMETRIC_PRIOR = {"source": "internal"}
    _AVAILABLE_GEOMETRIC_PRIOR = ["internal", "classif", "segm"]

    def __init__(self, img: xr.Dataset, **cfg: Union[str, int, float, bool]):
        """
        :param img: xarray.Dataset of image with metadata
        :type img: xarray.Dataset
        :param cfg: optional configuration, {}
        :type cfg: dict
        :return: None
        """
        super().__init__(img, **cfg)
        self.cfg = self.check_geometric_prior(cfg)
        self._geometric_prior = self.cfg["geometric_prior"]
        if self._geometric_prior["source"] == "classif":  # type: ignore
            self._classes = self._geometric_prior["classes"]  # type: ignore

    def check_geometric_prior(self, cfg: dict) -> dict:
        """
        Verify geometric_prior parameter for 3sgm classification
        :param cfg:
        :type cfg: dict
        :return cfg: optimization configuration updated
        :rtype cfg: dict
        """

        # Geometric prior is internal if not mentioned
        if "geometric_prior" not in cfg:
            cfg["geometric_prior"] = self._GEOMETRIC_PRIOR
        elif isinstance(cfg["geometric_prior"], dict):
            if not cfg["geometric_prior"]["source"] in self._AVAILABLE_GEOMETRIC_PRIOR:
                logging.error("%s is not available as a geometric prior", cfg["geometric_prior"]["source"])

        # Must be classif, segm or internal
        if cfg["geometric_prior"]["source"] not in self._AVAILABLE_GEOMETRIC_PRIOR:
            logging.error("%s is not available as a geometric prior source", cfg["geometric_prior"]["source"])
            sys.exit(1)

        # If source is classif, classes must be instantiated
        if cfg["geometric_prior"]["source"] == "classif":
            if "classes" not in cfg["geometric_prior"]:
                logging.error("Classes must be instantiated if source is classif")
                sys.exit(1)
            else:
                # If source is classif, classes must be instantiated as a list of string
                if not isinstance(cfg["geometric_prior"]["classes"], list) and not all(
                    isinstance(item, str) for item in cfg["geometric_prior"]["classes"]
                ):
                    logging.error("Classes must be a list of string")
                    sys.exit(1)

        if cfg["geometric_prior"]["source"] == "segm":
            if "classes" in cfg["geometric_prior"]:
                logging.error("Classes can't be instantiated if source is segm")
                sys.exit(1)

        return cfg

    def desc(self):
        """
        Describes the optimization method
        """
        print("Optimization with 3SGM")

    def compute_optimization_layer(
        self, cv: xr.Dataset, img_left: xr.Dataset, img_shape: Tuple[int, ...]
    ) -> np.ndarray:
        """
        Compute optimization layer for optimization method

        :param cv: the cost volume, with the data variables:

                - cost_volume 3D xarray.DataArray (row, col, disp)
                - confidence_measure 3D xarray.DataArray (row, col, indicator)
        :type cv: xarray.Dataset
        :param img_left: left Dataset image containing :

                - im : 2D (row, col) or 3D (band, row, col) xarray.DataArray
                - msk (optional): 2D (row, col) xarray.DataArray
        :type img_left: xarray
        :param img_shape: shape of the input image
        :type img_shape: Tuple[int, ...]
        :return: the optimization layer array
        :rtype: np.ndarray
        """
        # internal (from cv), segm or classif (from image)
        mode = self._geometric_prior["source"]  # type: ignore

        if mode in ["segm", "classif"]:
            # if geometric_prior comes from the image (segm or classif)
            if mode in img_left:
                geometric_prior_array = img_left[mode].data
                if mode == "classif":
                    geometric_prior_array = fuse_classification_bands(img_left, self._classes)
            else:
                logging.warning("%s not in image dataset.", mode)
                sys.exit(1)
        # if user wants to use another type of geometric prior
        else:
            geometric_prior_array = np.ones(img_shape)
            if mode != "internal":
                logging.warning(
                    "User wants to use a mode not in image dataset. \n "
                    "Default is used : no optimization with 3sgm will be performed."
                )
            else:
                if "internal" not in cv:
                    # if layer not computed we add a default one
                    prior_array = xr.DataArray(
                        data=geometric_prior_array,
                        coords=[("row", np.arange(img_shape[0])), ("col", np.arange(img_shape[1]))],
                    )
                    logging.warning(
                        "User wants to use a mode not in cost volume. \n "
                        "Default is used : no optimization with 3sgm will be performed."
                    )
                    logging.warning(
                        "If semantic_segmentation is present in the pipeline, \n "
                        "it should be placed before the 3SGM optimization step."
                    )
                    # Apply geometric prior to cost volume
                    cv["internal"] = prior_array
                else:
                    geometric_prior_array = cv["internal"].data

        geometric_prior_array = geometric_prior_array.astype(np.float32)
        geometric_prior_array[np.isnan(geometric_prior_array)] = -9999

        return geometric_prior_array
